"""
Name: Anthropic
URL: https://anthropic.ai/
Features:
- Chat
"""
import json
from typing import Union

import httpx
import requests

from ..chat.entities import (
    NOT_GIVEN,
    ChatConfig,
    IChatClient,
    LogConfig,
    MessageChunk,
    Prompt,
    Response,
    Stream,
)
from .connector import IConnector
from .providers import Providers


def _process_chunk(obj) -> MessageChunk:
    return MessageChunk(content=obj["completion"])


class AnthropicStream:
    def __init__(self, iterator):
        self.iterator = iterator
        self.closed = False

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            if self.closed:
                raise StopIteration

            obj = next(self.iterator)
            line = obj.decode("utf-8")
            if line.startswith("data: "):
                data = json.loads(line[6:])
                if data.get("stop_reason") is not None:
                    self.closed = True
                if data.get("completion") is not None:
                    return data
            else:
                continue


class AnthropicResponse(Response):
    def __init__(self, obj):
        content = obj["completion"]
        super().__init__(
            content=content,
            prompt_tokens=None,
            completion_tokens=None,
            raw=obj,
        )


class AnthropicConnector(IConnector, IChatClient):
    def __init__(
        self,
        api_key: str = None,
        client: "Speck" = None,
        log_config: "LogConfig" = None,
    ):
        super().__init__(
            client=client, provider=Providers.Anthropic, log_config=log_config
        )
        if api_key is not None:
            self.api_key = api_key
            self.url = "https://api.anthropic.com/v1/complete"

    def _convert_messages_to_prompt(self, prompt: Prompt) -> str:
        res = ""
        messages = prompt.compose()
        if messages[0]["role"] == "system":
            res = "System: " + messages[0]["content"] + "\n\n"
        for msg in messages:
            if msg["role"] == "system":
                continue
            res += (
                f"{'Human' if msg['role'] == 'user' else 'Assistant'}: "
                + msg["content"]
                + "\n\n"
            )
        res += "Assistant:"
        return res

    def _process_kwargs(self, prompt: Prompt, config: ChatConfig, **config_kwargs):
        if config is NOT_GIVEN:
            config = ChatConfig(**config_kwargs)
            # Todo: convert to default config based on class param

        # Remove all None values
        all_kwargs = {k: v for k, v in vars(config).items() if v is not None}

        input = self._convert_messages_to_prompt(prompt)

        headers = {
            "anthropic-version": config.get("anthropic_version", "2023-06-01"),
            "content-type": "application/json",
            "x-api-key": self.api_key,
        }

        data = {
            "model": config.model,
            "prompt": input,
            "max_tokens_to_sample": config.max_tokens or 100,
            "stream": config.stream,
            "temperature": config.temperature,
            **config.chat_args,
        }

        blocked_kwargs = ["provider", "_log", "chat_args", "stream", "max_tokens"]

        for k, v in all_kwargs.items():
            if k not in data and k not in blocked_kwargs:
                data[k] = v

        return headers, data, all_kwargs

    def chat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[AnthropicResponse, Stream]:
        headers, data, all_kwargs = self._process_kwargs(
            prompt, config, **config_kwargs
        )

        response = requests.post(
            self.url, headers=headers, data=json.dumps(data), stream=config.stream
        )

        if config.stream:
            return Stream(
                client=self._client,
                iterator=AnthropicStream(response.iter_lines()),
                log_config=self._log_config,
                kwargs=self._get_log_kwargs(prompt, None, **all_kwargs),
                processor=_process_chunk,
            )
        else:
            output = response.json()
            if config._log:
                self.log(
                    log_config=self._log_config,
                    prompt=prompt,
                    response=AnthropicResponse(output),
                    **all_kwargs,
                )
                # Todo: set config= as param

            return AnthropicResponse(output)

    async def achat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[AnthropicResponse, Stream]:
        headers, data, all_kwargs = self._process_kwargs(
            prompt, config, **config_kwargs
        )

        with httpx.AsyncClient() as client:
            response = await client.post(
                self.url, headers=headers, data=json.dumps(data), stream=config.stream
            )

            if config.stream:
                return Stream(
                    client=self._client,
                    iterator=AnthropicStream(response.iter_lines()),
                    log_config=self._log_config,
                    kwargs=self._get_log_kwargs(prompt, None, **all_kwargs),
                    processor=_process_chunk,
                )
            else:
                output = response.json()
                if config._log:
                    self.log(
                        log_config=self._log_config,
                        prompt=prompt,
                        response=AnthropicResponse(output),
                        **all_kwargs,
                    )
                    # Todo: set config= as param

                return AnthropicResponse(output)



from enum import Enum


class Providers(Enum):
    Anthropic = "Anthropic"
    AzureOpenAI = "AzureOpenAI"
    OpenAI = "OpenAI"
    CustomProvider = "CustomProvider"
    Replicate = "Replicate"


ProvidersList: list[str] = []
for provider in Providers:
    ProvidersList.append(provider.value.lower())



import datetime
import os
import threading
from typing import Any


def generate_metadata_dict() -> dict[str, Any]:
    metadata: dict[str, Any] = {
        "thread": threading.get_ident(),
        "process": os.getpid(),
        # "timestamp": datetime.datetime.now(datetime.UTC).strftime(
        #     "%Y-%m-%d %H:%M:%S.%f"
        # )[:-3],
        "timestamp": datetime.datetime.now().timestamp(),
    }
    return metadata



from __future__ import annotations

import re
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Iterator, Literal, Optional, Tuple, Union
import requests

from openai._types import NotGiven

# from dataclasses import dataclass
from pydantic import BaseModel, Extra

from ..chat.logger import ChatLogger
from ..util._live_debug import run_debug_websocket

NOT_GIVEN = None


class MissingVariables(ValueError):
    variables: list[str]
    id: str | None

    def __init__(self, variables: list[str], id: str | None):
        self.variables = variables
        self.id = id

    def __str__(self):
        return f"Missing variables {self.variables} for prompt \"{self.id}\""


class PromptNotFound(ValueError):
    id: str

    def __init__(self, id: str):
        self.id = id

    def __str__(self):
        return f"Prompt \"{self.id}\" not found."


class DuplicatePromptId(ValueError):
    id: str

    def __init__(self, id: str):
        self.id = id

    def __str__(self):
        return f"Prompt \"{self.id}\" already exists."


class PromptContext:
    api_key: str
    endpoint: str
    context_id: str
    path: str
    prompts: dict[str, Prompt]

    def __init__(self, api_key: str, endpoint: str, path: str):
        self.api_key = api_key
        self.endpoint = endpoint or "https://api.getspeck.ai"
        self.path = path
        self.prompts = {}
        self.context_id = uuid.uuid4().hex

    def add_prompt(self, id: str, prompt: Prompt):
        if id in self.prompts:
            raise DuplicatePromptId(id)
        self.prompts[id] = prompt

    def upload(self):
        if self.api_key:
            headers = {"X-API-Key": self.api_key}
            body = self.to_dict()

            response = requests.post(f"{self.endpoint}/logging/create/prompt", headers=headers, json=body)
            response.raise_for_status()
            return response.json()

    def get_key(self):
        return {
            "context_id": self.context_id,
            "path": self.path,
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.upload()

    def to_dict(self):
        return {
            "context_id": self.context_id,
            "path": self.path,
            "prompts": {
                name: prompt.to_dict() for name, prompt in self.prompts.items()
            },
            "metadata": {
                "timestamp": int(datetime.now().timestamp()),
            }
        }


class Value:
    type: Type
    value: Any = None

    class Type(Enum):
        ANY = "any"

    def __init__(self, type: Type, value: Any):
        self.type = type
        self.value = value


class Message:
    role: MessageRole
    content: str
    condition: Union[str, None] = None

    def __init__(
        self,
        role: MessageRole,
        content: str,
        condition: Union[str, None] = None,
    ):
        self.role = role
        self.content = content
        self.condition = condition

    def to_dict(self):
        if self.condition:
            return {
                "role": self.role,
                "content": self.content,
                "condition": self.condition,
            }
        else:
            return {"role": self.role, "content": self.content}

    def __str__(self):
        return f"Message(role={self.role}, content={self.content}, condition={self.condition})"


class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"  # Returns the key in curly braces as a string


class PromptChild:
    prompt: Prompt = None
    content: str = None
    inherit_variables: bool = True

    @property
    def is_prompt(self):
        return self.prompt is not None

    def __init__(self, prompt: Prompt= None, content: str = None, inherit_variables: bool = True):
        self.prompt = prompt
        self.content = content
        self.inherit_variables = inherit_variables

    def get_ref(self):
        if self.content is not None:
            return {
                "content": self.content,
                "inherit_variables": self.inherit_variables,
            }

        return {
            "prompt": self.prompt.get_ref(),
            "inherit_variables": self.inherit_variables,
        }

class Prompt(str):
    id: Union[str, None] = None
    messages: list[Message]
    variable_declarations: Union[dict[str, Value]] = None
    variables: Union[dict[str, Any], None] = None
    config: Union[dict[str, str], None] = None
    _context: PromptContext = None
    _children: list[PromptChild] = []

    def get_children_messages(self):
        if self._children is None:
            return None
        ret = []
        for child in self._children:
            if child.is_prompt:
                if child.inherit_variables:
                    raise NotImplementedError("Inheriting variables not supported yet.")
                ret += child.prompt.compose()
            else:
                ret.append(
                    {
                        "role": "user",
                        "content": Prompt._format_map(
                            child.content, self.variables, self.variable_declarations
                        ) if child.inherit_variables else child.content,
                    }
                )

        return ret

    def to_dict(self):
        messages = [msg.to_dict() for msg in self.messages]

        children_messages = self.get_children_messages()
        if children_messages:
            messages += children_messages

        return {
            "messages": messages,
            "variables": self.variables,
            "config": self.config,
            "id": self.id,
        }

    def get_ref(self):
        data = {
            **(self._context.get_key() if self._context else {}),
            "id": self.id,
            "variables": self.variables,
        }

        if self._children:
            data["children"] = [child.get_ref() for child in self._children]

        return data

    def __init__(
        self,
        messages: PromptTypes,
        variables: Union[dict[str, str], None] = None,
        config: Union[dict[str, str], None] = None,
        id: Union[str, None] = None,
        context: PromptContext = None,
        variable_declarations: Union[dict[str, Value], None] = None,
        children: list[PromptChild] = None,
    ):
        if isinstance(messages, str):
            messages = [Message(role="user", content=messages)]
        elif isinstance(messages, Message):
            messages = [messages]
        elif isinstance(messages, list):
            if all(isinstance(message, Message) for message in messages):
                pass
            elif all(isinstance(message, dict) for message in messages):
                messages = [
                    Message(role=message["role"], content=message["content"])
                    for message in messages
                ]
            else:
                raise ValueError(
                    f"Invalid type for messages: {type(messages)}\n{messages}"
                )

        self.messages = messages
        self.variables = variables
        self.config = config
        self.id = id
        self._context = context
        self.variable_declarations = variable_declarations
        self._children = children
        super().__init__()

    @classmethod
    def create(
        cls,
        messages: PromptTypes,
        variables: dict[str, str] = None,
        id: Union[str, None] = None,
        config: Union[dict[str, Any], None] = None,
    ) -> "Prompt":
        if isinstance(messages, cls):
            # Todo: clone object and add variables
            return messages
        return cls(messages=messages, variables=variables, id=id, config=config)

    @classmethod
    def _read(cls, lines: str, id: str = None, context: PromptContext = None) -> "Prompt":
        # Todo: add config parsing
        config = {}
        messages = []
        variable_declarations = {}

        current_min_spaces = 0
        current_section = None
        current_message = []
        condition = None

        def add_message():
            nonlocal current_message, current_min_spaces, condition, current_section
            if current_section:  # current_message:
                if current_section != "variables" and not current_section.startswith("/") and current_section != "config":
                    messages.append(
                        Message(
                            role=current_section,
                            content="\n".join(
                                [m[current_min_spaces:] for m in current_message]
                            ),
                            condition=condition,
                        )
                    )
                current_message = []
                current_min_spaces = 0
                condition = None

        for line in lines.split("\n"):
            line = line.rstrip("\r")
            if line.startswith("<"):
                line = line.strip()
                add_message()
                current_section = line[1:-1]
                if "if " in current_section:
                    if_index = current_section.index("if ")
                    condition = current_section[if_index + 2 :]
                    current_section = current_section[: if_index - 1]
                else:
                    condition = None
            elif current_section == "config" and "=" in line:
                key, value = line.split("=", 1)
                config[key.strip()] = value.strip()
            elif current_section == "variables":
                var_value = None
                if "=" in line:
                    line, var_value = line.split("=", 1)
                    line = line.strip()
                    var_value = var_value.strip()

                var_type = Value.Type.ANY
                if ":" in line:
                    line, var_type = line.split(":", 1)
                    line = line.strip()
                    var_type = var_type.strip()

                key = line.strip()
                variable_declarations[key] = Value(type=var_type, value=var_value)
            elif (
                current_section is not None
                and current_section in ["system", "user", "assistant", "if", "else"]
                or "add " in current_section
            ):
                min_spaces = len(line) - len(line.lstrip())
                if 0 < min_spaces < current_min_spaces or current_min_spaces == 0:
                    current_min_spaces = min_spaces
                current_message.append(line)

        add_message()
        return cls(messages=messages, id=id, context=context, variable_declarations=variable_declarations, config=config)

    @classmethod
    def read_all(cls, path: str, context: PromptContext = None) -> dict[str, "Prompt"]:
        with open(path, "r") as f:
            return cls.reads_all(f.read(), context=context)

    @classmethod
    def read(cls, path: str, id: Union[str, None] = None, context: PromptContext = None, **kwargs) -> "Prompt":
        with open(path, "r") as f:
            return cls.reads(f.read(), id=id, context=context, **kwargs)

    @classmethod
    def reads(cls, text: str, id: Union[str, None] = None, context: PromptContext = None, **kwargs) -> "Prompt":
        name = id or kwargs.get("name")
        if name is not None:
            prompts = cls.reads_all(text, context=context)
            return prompts[name]
        else:
            return cls._read(text, id=name)

    @classmethod
    def parse(cls, text: str, context: PromptContext = None) -> dict[str, "Prompt"]:
        return cls.reads_all(text, context=context)

    @classmethod
    def reads_all(cls, text: str, context: PromptContext = None) -> dict[str, "Prompt"]:
        if context is None:
            context = PromptContext(path=text.split("\n")[0], api_key=None, endpoint=None)

        prompts = {}
        lines = []
        current_prompt_name = None
        current_min_spaces = -1

        for line in text.split("\n"):
            line = line.rstrip("\n").rstrip("\r")
            if line.lstrip().startswith("<"):
                min_spaces = len(line) - len(line.lstrip())
                stripped_line = line.strip()

                if stripped_line.startswith("<prompt") and min_spaces == 0:
                    if current_prompt_name:
                        prompts[current_prompt_name] = cls._read(
                            "\n".join([m[current_min_spaces:] for m in lines]),
                            id=current_prompt_name,
                            context=context,
                        )

                    current_min_spaces = -1
                    lines = []
                    temp_prompt_name = stripped_line[8:-1].strip()
                    if "id=" in temp_prompt_name:
                        temp_prompt_name, id = temp_prompt_name.split("id=")
                        current_prompt_name = id.strip("\"' ")
                    else:
                        current_prompt_name = temp_prompt_name
                elif stripped_line.startswith("</prompt>") and min_spaces == 0:
                    prompts[current_prompt_name] = cls._read(
                        "\n".join([m[current_min_spaces:] for m in lines]),
                        id=current_prompt_name,
                        context=context,
                    )
                    current_prompt_name = None
                    current_min_spaces = -1
                    lines = []
                else:
                    lines.append(line)
                    if current_min_spaces == -1 or min_spaces < current_min_spaces:
                        current_min_spaces = min_spaces
            else:
                lines.append(line)

        for id, prompt in prompts.items():
            context.add_prompt(id, prompt)

        return prompts

    def _file(self):
        file = []
        for message in self.messages:
            file.append(f"<{message.role.lower()}>")
            for line in message.content.split("\n"):
                file.append(" " * 4 + line)
        return "\n".join(file)

    @classmethod
    def write(cls, prompt: Union["Prompt", dict[str, "Prompt"]], path: str):
        with open(path, "w") as f:
            if isinstance(prompt, dict):
                content = ""
                for name, prompt in prompt.items():
                    content += f'<prompt id="{name}">\n'
                    content += "\n".join(
                        [" " * 4 + line for line in prompt._file().split("\n")]
                    )
                    content += "\n</prompt>\n\n"
                f.write(content.strip())
            else:
                f.write(prompt._file())

    def __new__(
        cls,
        messages: PromptTypes,
        **kwargs,
    ):
        # Todo: Handle string, Message, and list[Message]
        instance = super(Prompt, cls).__new__(cls, str(messages))
        return instance

    @classmethod
    def from_openai(cls, messages: list[dict[str, str]]):
        return cls(
            messages=[
                Message(role=message["role"], content=message["content"])
                for message in messages
            ]
        )

    def _check_condition(self, condition: str):
        if condition is None:
            return False

        condition = condition.strip()
        if condition.startswith("not"):
            return not self._check_condition(condition[3:])
        if condition.startswith("!"):
            return not self._check_condition(condition[1:])

        def check_declarations():
            if self.variable_declarations is None:
                return False

            var_name = condition.strip(" {}")
            if var_name not in self.variable_declarations:
                return False

            value = self.variable_declarations[var_name].value
            return (
                value == "true"
                or value == "True"
                or value == "1"
                or value == 1
                or value == True
            )

        def check_variables():
            if self.variables is None:
                return check_declarations()

            var_name = condition.strip(" {}")
            if var_name not in self.variables:
                return check_declarations()

            value = self.variables[var_name]
            return (
                value == "true"
                or value == "True"
                or value == "1"
                or value == 1
                or value == True
            )

        return check_variables()

    @staticmethod
    def _format_map(text: str, variables: dict[str, str], variable_declarations: dict[str, Value]):
        pattern = r"\{[^{}]*\}"
        matches = list(re.finditer(pattern, text))
        for match in reversed(matches):
            start, end = match.span()
            var_text = text[start + 1 : end - 1]
            if variables is not None and var_text in variables:
                # Todo: Promptfiles better variable type parsing
                text = text[:start] + str(variables[var_text]) + text[end:]
            elif variable_declarations is not None and var_text in variable_declarations:
                text = text[:start] + str(variable_declarations[var_text].value) + text[end:]

        return text  # string.format_map(SafeDict(variables or {}))

    def _convert_expression(self, expression: str) -> str:
        """In (A ? B : C), B and C can be expressions"""
        expression = expression.strip()

        if expression.startswith('"') and expression.endswith('"'):
            return Prompt._format_map(
                expression[1:-1].replace("\\n", "\n").replace("\\t", "\t"),
                self.variables,
                self.variable_declarations,
            )
        else:
            return Prompt._format_map("{" + expression + "}", self.variables, self.variable_declarations)

    def _format_message(self, message: Message, override_role: Union[str, None] = None):
        # Check for variables with {...: prompt} and replace with prompt
        pattern = r"\{[^{}]*\}"  # Pattern to match {anything: prompt}

        text = message.content
        matches = list(re.finditer(pattern, text))
        # Replace matches from the end to keep indexes correct
        for match in reversed(matches):
            start, end = match.span()
            var_text = text[start + 1 : end - 1]
            # Adds support for ternary operators and conditions
            if "?" in var_text:
                condition, values = var_text.split("?")
                left, right = values.split(":")
                if self._check_condition(condition):
                    text = text[:start] + self._convert_expression(left) + text[end:]
                else:
                    text = text[:start] + self._convert_expression(right) + text[end:]
            # text = text[:start] + "{hi}" + text[end:]

        return {
            "role": override_role if override_role is not None else message.role,
            "content": Prompt._format_map(text, self.variables, self.variable_declarations),
        }

    def compose(self, override_prompt: Prompt = None) -> list[dict[str, str]]:
        var_declarations = (self.variable_declarations if override_prompt is None else override_prompt.variable_declarations) or {}
        required_variables = set(var for var, value in var_declarations.items() if value.value is None)
        defined_variables = set(self.variables or {})

        prev_role, prev_condition = None, False
        ret: list[dict[str, str]] = []

        def add_prompt(name: str):
            if self._context and name in self._context.prompts:
                for sub_msg in self.compose(
                    override_prompt=self._context.prompts[name]
                ):
                    if sub_msg["role"] == "system":
                        raise NotImplementedError("System prompt not supported yet.")
                    else:
                        ret.append(sub_msg)
            else:
                raise PromptNotFound(name)

        for message in (
            self.messages if override_prompt is None else override_prompt.messages
        ):
            add = False  # Java-like variable scoping
            if message.condition is not None:
                add = self._check_condition(message.condition)
                prev_role, prev_condition = "if", add
            elif message.role == "else":
                if not prev_condition:
                    if message.role.startswith("else add "):
                        id = message.role[9:].strip()
                        if "id=" in id:
                            id = id.split("id=")[1].strip("\"' ")
                        add_prompt(id)
                    else:
                        ret.append(
                            self._format_message(message, override_role=prev_role)
                        )
                prev_role, prev_condition = "else", True
                continue
            else:
                add = True
                prev_role = message.role

            if add:
                if message.role.startswith("add "):
                    id = message.role[4:].strip()
                    if "id=" in id:
                        id = id.split("id=")[1].strip("\"' ")
                    add_prompt(id)
                else:
                    ret.append(self._format_message(message))

        children_messages = self.get_children_messages()
        if children_messages:
            ret += children_messages

        if len(required_variables - defined_variables) > 0:
            raise MissingVariables(variables=list(required_variables - defined_variables), id=(override_prompt or self).id)

        return ret

    def add(self, role: str = None, content: str = None, prompt: Prompt = None, inherit_variables=False) -> Prompt:
        new_prompt = self.__class__(
            messages=self.messages,
            variables=self.variables,
            id=self.id,
            config=self.config,
            variable_declarations=self.variable_declarations,
            context=self._context,
            children=(self._children or []),
        )

        if role is not None or content is not None:
            new_prompt._children.append(PromptChild(content=content))

        if prompt is not None:
            new_prompt._children.append(PromptChild(prompt=prompt, inherit_variables=inherit_variables))

        return new_prompt

    @staticmethod
    def _apply_variables(
        messages: list[Message], variables: dict[str, str], variable_declarations: dict[str, Value]
    ) -> list[Message]:
        return [
            Message(
                role=message.role,
                content=Prompt._format_map(message.content, variables, variable_declarations),
                condition=message.condition,
            )
            for message in messages
        ]

    def _check_duplicate_keys(self, other_variables: dict[str, str]) -> dict[str, str]:
        duplicate_keys = set((self.variables or {}).keys()).intersection(
            set((other_variables or {}).keys())
        )
        return {
            key: self.variables[key]
            for key in duplicate_keys
            if self.variables[key] != other_variables[key]
        }

    def _remove_duplicate_keys_from_messages(
        self, other_variables: dict[str, str]
    ) -> list[Message]:
        messages = self.messages
        applied_variables = self._check_duplicate_keys(other_variables)
        if len(applied_variables) > 0:
            messages = self._apply_variables(self.messages, applied_variables, None)

        return messages

    def format(self, *args, **kwargs) -> Prompt:
        # return self.__class__(
        #     messages=[
        #         Message(
        #             role=message.role, content=message.content.format(*args, **kwargs)
        #         )
        #         for message in self.messages
        #     ]
        # )

        messages = self._remove_duplicate_keys_from_messages(kwargs)
        return self.__class__(
            id=self.id,
            messages=[
                Message(
                    role=message.role,
                    content=message.content,
                    condition=message.condition,
                )
                for message in messages
            ],
            variables={**SafeDict(self.variables or {}), **kwargs},
            context=self._context,
            config=self.config,
            variable_declarations=self.variable_declarations,
            children=self._children,
        )

    def __add__(self, other):
        if isinstance(other, Message):
            return self.__class__(
                messages=self.messages + [other], variables={**(self.variables or {})}
            )
        elif isinstance(other, Prompt):
            # Check if there are duplicate keys
            messages = self._remove_duplicate_keys_from_messages(other.variables or {})

            return self.__class__(
                messages=messages + other.messages,
                variables={
                    **SafeDict(self.variables or {}),
                    **SafeDict(other.variables or {}),
                },
            )
        else:
            raise NotImplementedError

    def __str__(self):
        return (
            "Prompt(\n"
            + (
                "\n".join(
                    [f"{message.role}: {message.content}" for message in self.messages]
                )
                + "\n"
                + str(self.variables or {})
            )
            + "\n)"
        )


class Response(BaseModel):
    content: Union[str, None]
    prompt_tokens: Union[int, None] = None
    completion_tokens: Union[int, None] = None
    raw: Union[dict, None] = None

    def __init__(
        self,
        content: Union[str, None],
        closed: bool = False,
        prompt_tokens: Union[int, None] = None,
        completion_tokens: Union[int, None] = None,
        raw: Union[dict, None] = None,
        **kwargs,
    ):
        super().__init__(
            content=content,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            raw=raw,
        )
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def create(cls, response: ResponseTypes) -> "Response":
        if isinstance(response, cls):
            return response
        elif isinstance(response, str):
            return cls(content=response)
        else:
            raise NotImplementedError

    def __str__(self):
        return f"Response(content={self.content}, prompt_tokens={self.prompt_tokens}, completion_tokens={self.completion_tokens}, raw={self.raw})"


class MessageChunk(BaseModel):
    content: Union[str, None]

    def encode(self, encoding: str = "utf-8"):
        content = self.content or ""
        return content.encode(encoding)


class Stream:
    # processor that has lambda which returns MessageDelta
    def __init__(
        self,
        client: "Speck",
        iterator: Iterator[Any],
        kwargs: dict,
        log_config: "LogConfig",
        processor: Callable[[Any], MessageChunk],
        custom_metadata: dict = None,
    ):
        self._client = client
        self.message: str = ""
        self.tokens: int = 0
        self._iterator = iterator
        self._kwargs = kwargs
        self._processor = processor
        self._has_logged = False
        self._closed = False
        self._log_config = log_config
        self.custom_metadata = custom_metadata

    def _log(self):
        if not self._has_logged:
            self._has_logged = True

            kwargs = self._kwargs
            kwargs["prompt"] = self._kwargs.get("prompt", [])
            kwargs["temperature"] = self._kwargs.get("temperature", "N/A")
            kwargs["model"] = self._kwargs.get("model", "N/A")
            kwargs["response"] = Response(
                content=self.message, raw={}, closed=True, completion_tokens=self.tokens
            )

            # Todo: add prompt_tokens using tiktoken
            ChatLogger.log(log_config=self._log_config, custom_metadata=self.custom_metadata, **kwargs)

    def _process(self, item) -> MessageChunk:
        return self._processor(item)

    def __next__(self) -> MessageChunk:
        try:
            if self._closed:
                raise StopIteration

            # next_item = None
            # while next_item is None:
            next_item = next(self._iterator)

            item: MessageChunk = self._process(next_item)
            if item.content:
                self.message += item.content
            self.tokens += 1
            return item
        except StopIteration:
            self._log()
            raise

    def __iter__(self) -> Iterator[MessageChunk]:
        return self

    def close(self):
        try:
            self._closed = True
            # todo: make this work for packages other than openai
            self._iterator.response.close()
        except AttributeError:
            pass


class LogConfig(BaseModel):
    api_key: Union[str, None] = None
    endpoint: str = "https://api.getspeck.ai"

    class Config:
        extra = "allow"


class ChatConfig:
    # Todo: add typed params here
    # Todo: Create conversions to other formats
    def __init__(
        self,
        *,
        provider: str = None,
        model: OpenAIModel,
        stream: bool = False,
        _log: bool = True,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        max_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
        frequency_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        presence_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        chat_args: dict[str, str] = {},
        **config_kwargs,
    ):
        if "log_config" in config_kwargs:
            del config_kwargs["log_config"]

        self.provider = provider
        self.model = model
        self.stream = stream
        self._log = _log
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.chat_args = chat_args | config_kwargs
        # If this is modified, update to_dict too

    def to_dict(self):
        return {
            "provider": self.provider,
            "model": str(self.model),  # Assuming model can be represented as a string
            "stream": self.stream,
            "_log": self._log,
            "temperature": self._convert_optional(self.temperature),
            "max_tokens": self._convert_optional(self.max_tokens),
            "top_p": self._convert_optional(self.top_p),
            "frequency_penalty": self._convert_optional(self.frequency_penalty),
            "presence_penalty": self._convert_optional(self.presence_penalty),
            "chat_args": self.chat_args,
        }

    def _convert_optional(self, value):
        return None if isinstance(value, NotGiven) else value

    @classmethod
    def create(cls, config: ChatConfigTypes, kwargs: dict = None) -> "ChatConfig":
        if isinstance(config, cls):
            if kwargs is not None:
                return cls(**{**config.__dict__, **kwargs})
            else:
                return config
        elif isinstance(config, dict):
            return cls(**config)
        elif kwargs:
            return cls(**kwargs)
        else:
            raise NotImplementedError

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def convert(self, provider: str = "speck") -> "ChatConfig":
        """
        Convert to another config format
        """
        if provider == "openai":
            return OpenAIChatConfig(
                model=self.model,
                stream=self.stream,
                _log=self._log,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                **self.chat_args,
            )

        return self

    def log_chat(
        self,
        *,
        log_config: LogConfig,
        prompt: Prompt,
        response: Response,
        provider: str = "speck",
        prompt_ref: dict = None,
        custom_metadata: dict = None,
    ):
        config = self.convert()
        ChatLogger.log(
            log_config=log_config,
            provider=provider,
            model=str(config.model),
            prompt=prompt,
            response=response,
            prompt_ref=prompt_ref,
            custom_metadata=custom_metadata,
            **config.chat_args,
        )

    def encode(self, encoding: str = "utf-8"):
        return self.__str__().encode(encoding)

    def __str__(self):
        return f"ChatConfig(provider={self.provider}, model={self.model}, stream={self.stream}, _log={self._log}, temperature={self.temperature}, max_tokens={self.max_tokens}, top_p={self.top_p}, frequency_penalty={self.frequency_penalty}, presence_penalty={self.presence_penalty}, chat_args={self.chat_args})"


class OpenAIChatConfig(ChatConfig):
    def __init__(
        self,
        model: OpenAIModel,
        provider: str = "openai",
        stream: bool = False,
        _log: bool = True,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        max_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
        frequency_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        presence_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        **config_kwargs,
    ):
        self.provider = provider
        self.model = model
        self.stream = stream
        self._log = _log
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.chat_args = config_kwargs

    def convert(self, provider: str = "speck") -> ChatConfig:
        """
        Maps config to universal format then converts to another config format
        """
        universal_config = ChatConfig(
            model=self.model,
            stream=self.stream,
            _log=self._log,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            **self.chat_args,
        )

        return universal_config.convert(provider=provider)


class IChatClient(ABC):
    def debug_chat(
        self, prompt: "Prompt", config: "ChatConfig"
    ) -> ("Prompt", "ChatConfig"):
        data = run_debug_websocket(self._client, self, prompt, config)

        if data.get("prompt") and data.get("config"):
            prompt = Prompt(**data["prompt"])
            config = ChatConfig(**data["config"])

        return prompt, config

    @abstractmethod
    def chat(
        self,
        prompt: PromptTypes,
        config: Union[ChatConfig, NotGiven] = NOT_GIVEN,
        **config_kwargs,
    ) -> Union[Response, Stream]:
        pass

    @abstractmethod
    async def achat(
        self,
        prompt: PromptTypes,
        config: Union[ChatConfig, NotGiven] = NOT_GIVEN,
        **config_kwargs,
    ) -> Union[Response, Stream]:
        pass


PromptTypes = Union[str, Message, list[Message], list[dict[str, str]]]
ResponseTypes = Union[Response, str]
ChatConfigTypes = Union[ChatConfig, dict[str, str]]

MessageRole = Union[Literal["system", "user", "assistant"], str]
OpenAIModel = Union[Literal["gpt-4", "gpt-3.5", "gpt-3.5-turbo"], str]



from typing import Any

from ..logs.logger import universal_format_log


class ChatLogger:
    @staticmethod
    def log(log_config: "LogConfig", prompt: Any, model: str, response: Any, prompt_ref: dict = None, custom_metadata: dict = None, **kwargs):
        universal_format_log(
            log_config=log_config,
            prompt=prompt,
            model=model,
            response=response,
            prompt_ref=prompt_ref,
            custom_metadata=custom_metadata,
            **kwargs,
        )



import inspect


def filter_kwargs(func, all_kwargs, exclude=[]):
    func_signature = inspect.signature(func)
    valid_params = set(func_signature.parameters)
    filtered_kwargs = {
        k: v for k, v in all_kwargs.items() if k in valid_params and k not in exclude
    }
    return filtered_kwargs


def no_filter_kwargs(func, all_kwargs, exclude=[]):
    filtered_kwargs = {k: v for k, v in all_kwargs.items() if k not in exclude}
    return filtered_kwargs



from enum import Enum
from typing import Union

from ..chat.entities import IChatClient, Prompt, Response, Stream
from ..connections.custom import CustomProviderConnector
from ..connections.openai import OpenAIConnector
from ..connections.providers import Providers
from ..connections.replicate import ReplicateConnector


class Formats(Enum):
    OpenAI = "openai"
    Replicate = "replicate"


class ChatClient(IChatClient):
    def __init__(self, provider_config: dict = None, **data):
        # Todo: refactor this legacy class
        self.provider_config = provider_config or {}
        self.connector = self._get_connector(**data)
        # Todo: add support for this
        self.format: Formats = Formats.OpenAI

    def _get_connector(self, **data) -> IChatClient:
        if self.provider == Providers.OpenAI:
            return OpenAIConnector(
                api_key=self.provider_config.get("api_key", data.get("api_key", ""))
            )
        elif self.provider == Providers.CustomProvider:
            return CustomProviderConnector(
                message_prefix=self.provider_config.get("message_prefix", ""),
                message_suffix=self.provider_config.get("message_suffix", ""),
            )
        elif self.provider == Providers.Replicate:
            return ReplicateConnector()
        raise NotImplementedError("Provider not supported")

    @classmethod
    def from_string(cls, model_str: str):
        provider_str, model_str = model_str.split(":")
        provider = next((p for p in Providers if p.value[0] == provider_str), None)
        if provider is None:
            raise ValueError("Invalid provider")
        return cls(provider=provider)

    @classmethod
    def from_openai(cls, api_key: str) -> OpenAIConnector:
        # return cls(provider=Providers.OpenAI, provider_config={"api_key": api_key})
        return OpenAIConnector(api_key=api_key)

    @classmethod
    def from_replicate(cls, api_key: Union[str, None] = None) -> ReplicateConnector:
        """Reads api_key from environment variable if not provided"""
        return ReplicateConnector(api_key=api_key)

    def chat(
        self, messages: Prompt, model: str, **config_kwargs
    ) -> Union[Response, Stream]:
        return self.connector.chat(messages=messages, model=model, **config_kwargs)



import requests

from .metadata import generate_metadata_dict


class app:
    api_key: str = None

    @classmethod
    def set_api_key(cls, api_key: str) -> None:
        cls.api_key = api_key

    @classmethod
    def log(
        cls, message: str, endpoint: str = "https://api.getspeck.ai"
    ) -> dict[str, str]:
        if not cls.api_key:
            return {}

        body: dict[str, str] = {
            "message": message,
            "metadata": generate_metadata_dict(),
        }

        headers = {"X-API-Key": cls.api_key}
        request: requests.Response = requests.post(
            f"{endpoint}/logging/create/app", headers=headers, json=body
        )
        return request.json()



import warnings
from typing import Any

import requests
import logging

from ..util import get_dict
from .metadata import generate_metadata_dict

logger = logging.getLogger(__name__)


# Todo: Fix typing for this function (circular import)
def universal_format_log(
    log_config: "LogConfig",
    provider: "Providers",
    prompt: "Prompt",
    model: str,
    response: "Response",
    session_key: str = None,
    prompt_ref: dict = None,
    custom_metadata: dict = None,
    **kwargs,
) -> dict[str, str]:
    if not log_config:
        logger.info("No log config found. Define the log config in the log or client.")
        return {}
    if not log_config.api_key:
        # print("No api key found. Define the api key in the log config or client.")
        return {}

    body: dict[str, Any] = {
        "input": {
            # Todo: Fix typing for Providers enum (circular import)
            "prompt_ref": prompt_ref,
            "provider": provider.value if hasattr(provider, "value") else provider,
            "model": model,
            **prompt.to_dict(),
            **kwargs,
        },
        "output": get_dict(response),
        "metadata": generate_metadata_dict(),
    }

    if custom_metadata is not None:
        body["custom_metadata"] = custom_metadata

    try:
        headers = {"X-API-Key": log_config.api_key}
        request: requests.Response = requests.post(
            f"{log_config.endpoint}/logging/create/llm", headers=headers, json=body
        )
        request.raise_for_status()
        return request.json()
    except requests.exceptions.HTTPError as e:
        logger.info("HTTP", e)
    except requests.exceptions.ReadTimeout as e:
        logger.info("Read", e)
    except requests.exceptions.ConnectionError as e:
        logger.info("Connection", e)
    except requests.exceptions.RequestException as e:
        logger.info("Request", e)



import functools
import inspect
import logging
from types import MethodType

logger = logging.getLogger(__name__)


def wrap_method(class_instance, method_name, wrapper):
    """
    Applies a wrapper function to a method of a class instance.

    Args:
        class_instance: The instance of the class whose method is to be wrapped.
        method_name (str): The name of the method to wrap.
        wrapper (function): The wrapper function to apply.

    Raises:
        AttributeError: If the specified method does not exist in the class instance.
        TypeError: If the specified attribute is not a method.
    """
    if not hasattr(class_instance, method_name):
        raise AttributeError(
            f"Method {method_name} not found in {class_instance.__class__.__name__}"
        )

    original_method = getattr(class_instance, method_name)
    # if not isinstance(original_method, MethodType):
    #     raise TypeError(
    #         f"Attribute {method_name} of {class_instance.__class__.__name__} is not a method"
    #     )

    @functools.wraps(original_method)
    def wrapped_method(*args, **kwargs):
        return wrapper(original_method, *args, **kwargs)

    setattr(class_instance, method_name, wrapped_method)


def add_method_arg(class_instance, method_name, new_arg_name, new_arg_default=None):
    """
    Dynamically adds a new positional argument to a method of a class instance.

    Args:
        class_instance: The instance of the class whose method is to be modified.
        method_name (str): The name of the method to modify.
        new_arg_name (str): The name of the new argument to add.
        new_arg_default: The default value of the new argument (None if not specified).
    """
    if not hasattr(class_instance, method_name):
        raise AttributeError(
            f"Method {method_name} not found in {class_instance.__class__.__name__}"
        )

    method = getattr(class_instance, method_name)
    if not isinstance(method, MethodType):
        raise TypeError(
            f"Attribute {method_name} of {class_instance.__class__.__name__} is not a method"
        )

    original_signature = inspect.signature(method)

    # Create a new parameter and add it to the original method's signature
    new_parameter = inspect.Parameter(
        name=new_arg_name,
        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        default=new_arg_default,
    )
    new_signature = original_signature.replace(
        parameters=[new_parameter, *original_signature.parameters.values()]
    )

    # Update the method's signature
    method.__signature__ = new_signature


def add_method_kwarg(
    class_instance, method_name, new_kwarg_name, new_kwarg_default=None
):
    """
    Dynamically adds a new keyword argument to a method of a class instance.

    Args:
        class_instance: The instance of the class whose method is to be modified.
        method_name (str): The name of the method to modify.
        new_kwarg_name (str): The name of the new keyword argument to add.
        new_kwarg_default: The default value of the new keyword argument.
    """
    if not hasattr(class_instance, method_name):
        raise AttributeError(
            f"Method {method_name} not found in {class_instance.__class__.__name__}"
        )

    method = getattr(class_instance, method_name)
    # if not isinstance(method, MethodType):
    #     raise TypeError(
    #         f"Attribute {method_name} of {class_instance.__class__.__name__} is not a method"
    #     )

    original_signature = inspect.signature(method)

    # Create a new keyword parameter and add it to the original method's signature
    new_parameter = inspect.Parameter(
        name=new_kwarg_name,
        kind=inspect.Parameter.KEYWORD_ONLY,
        default=new_kwarg_default,
    )
    new_signature = original_signature.replace(
        parameters=[*original_signature.parameters.values(), new_parameter]
    )

    # Update the method's signature
    method.__signature__ = new_signature



import asyncio
import json

import websockets


async def _websocket_client(client, connector, prompt, config):
    async with websockets.connect(client.ws_endpoint) as websocket:
        data = json.dumps(
            {
                "prompt": prompt.to_dict(),
                "config": config.to_dict(),
                "speck": client.to_dict(),
            }
        )
        # "connector": client.connector.to_dict()
        await websocket.send(data)

        try:
            print(
                f"[speck] LLM call ({config.provider}:{config.model}) sent to https://getspeck.ai/dash/debug"
            )
            # Wait for a response with a timeout of 5 seconds
            response = await asyncio.wait_for(websocket.recv(), timeout=1800)
            # print(f"Received from server: {json.loads(response)}")
            return json.loads(response)
        except asyncio.TimeoutError:
            # print("No response received from server within 5 seconds")
            pass
    return {}


def run_debug_websocket(client, connector, prompt, config):
    call = _websocket_client(client, connector, prompt, config)

    try:
        return asyncio.run(call)
    except RuntimeError:
        # Jupyter notebook async support
        try:
            import nest_asyncio

            nest_asyncio.apply()

            loop = asyncio.get_running_loop()
            task = loop.create_task(call)
            loop.run_until_complete(task)
            return task.result()
        except ImportError:
            print(
                "Jupyter requires nest_asyncio to be installed. Please install it with `pip install nest_asyncio`"
            )
            return {}


if __name__ == "__main__":
    # Call the synchronous function
    run_debug_websocket(None, None)



from abc import ABC

from ..chat.entities import ChatLogger, LogConfig, Prompt, Response
from .providers import Providers


class IConnector(ABC):
    _client: "Speck"
    _log_config: "LogConfig"

    def __init__(
        self, client: "Speck", provider: Providers, log_config: "LogConfig" = None
    ):
        self._client = client
        self.provider = provider
        self._log_config = log_config

    # @abstractmethod
    # def process_message(self, messages: Messages, model: str) -> str:
    #     pass

    def _get_log_kwargs(self, prompt: Prompt, response: Response, **kwargs):
        return {
            "provider": self.provider,
            "model": kwargs.get("model"),
            "temperature": kwargs.get("temperature"),
            "stream": kwargs.get("stream", False),
            "prompt": prompt,
            "config": kwargs,
            "response": response,
        }

    def log(
        self, *, log_config: LogConfig, prompt: Prompt, response: Response, custom_metadata: dict = None, **kwargs
    ):
        # Todo: refactor to use config.log_chat !!!
        ChatLogger.log(
            log_config=log_config,
            custom_metadata=custom_metadata,
            **self._get_log_kwargs(prompt, response, **kwargs),
        )

    def __str__(self):
        return f"Client({self.provider.value})"



"""
Name: Replicate
URL: https://replicate.com/
Features:
- Chat
"""
from typing import Union

import replicate
from replicate import Client

from ..chat.entities import (
    ChatConfig,
    IChatClient,
    LogConfig,
    MessageChunk,
    Prompt,
    Response,
    Stream,
)
from .connector import IConnector
from .providers import Providers

NOT_GIVEN = None


def _process_chunk(obj) -> MessageChunk:
    return MessageChunk(
        content=obj.data if str(obj.event) == "EventType.OUTPUT" else ""
    )


class ReplicateConnector(IConnector, IChatClient):
    """
    https://replicate.com/
    """

    def __init__(
        self,
        client: "Speck" = None,
        api_key: Union[str, None] = None,
        message_prefix: str = "<|im_start|>{role}\n",
        message_suffix: str = "<|im_end|>\n",
        messages_end: str = "<|im_start|>assistant\n",
        log_config: "LogConfig" = None,
    ):
        # Todo: support custom replicate model mappings
        # By default, built for meta/llama-2-70b
        super().__init__(
            client=client, provider=Providers.Replicate, log_config=log_config
        )
        self.api_key = api_key
        self.message_prefix = message_prefix
        self.message_suffix = message_suffix
        self.messages_end = messages_end
        self.replicate_client = Client(api_token=api_key)

    def _convert_messages_to_prompt(
        self, prompt: Prompt
    ) -> (str, list[dict[str, str]]):
        # Return tuple of system_prompt, prompt
        messages = prompt.compose()
        system_prompt = next(
            msg["content"] for msg in messages if msg["role"] == "system"
        )
        prompt = "".join(msg["content"] for msg in messages if msg["role"] != "system")
        return system_prompt, prompt

    def _process_args(self, prompt: Prompt, config: ChatConfig, **config_kwargs):
        # all_kwargs = {
        #     **config_kwargs,
        #     "temperature": max(temperature, 0.01) if temperature is not None else None,
        #     "max_new_tokens": max_tokens,
        #     "top_p": top_p,
        #     "repetition_penalty": max(presence_penalty, 0.01)
        #     if presence_penalty
        #     else None,
        #     "top_k": top_k,
        #     "test": test,
        # }
        blocked_kwargs = ["provider", "_log", "chat_args", "stream", "max_tokens"]
        mapped_args = {"repetition_penalty": "presence_penalty"}
        all_kwargs = {
            mapped_args.get(k, k): v
            for k, v in vars(config).items()
            if v is not None and k not in blocked_kwargs
        }
        if all_kwargs.get("temperature") < 0.01:
            all_kwargs["temperature"] = 0.01

        log_kwargs = {k: v for k, v in vars(config).items() if v is not None}

        # input = (
        #     "".join(
        #         self.message_prefix.format(role=msg.role)
        #         + msg.content
        #         + self.message_suffix.format(role=msg.role)
        #         for msg in prompt.messages
        #     )
        #     + self.messages_end
        # )
        system_prompt, user_prompt = self._convert_messages_to_prompt(prompt)
        all_kwargs["system_prompt"] = system_prompt

        # Remove unused kwargs (kept in, these values can cause errors)
        for arg in ["stream", "provider", "_log", "chat_args"]:
            if arg in all_kwargs:
                del all_kwargs[arg]

        return system_prompt, user_prompt, all_kwargs, log_kwargs

    def chat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[Response, Stream]:
        (
            system_prompt,
            user_prompt,
            all_kwargs,
            log_kwargs,
        ) = self._process_args(prompt, config, **config_kwargs)

        if config.stream:
            output = self.replicate_client.stream(
                config.model,
                input={"prompt": user_prompt, **all_kwargs},
            )

            return Stream(
                client=self._client,
                iterator=output,
                log_config=self._log_config,
                kwargs=self._get_log_kwargs(prompt, None, **log_kwargs),
                processor=_process_chunk,
            )
        else:
            output = self.replicate_client.run(
                config.model,
                input={"prompt": user_prompt, **all_kwargs},
            )

            content = "".join(output)

            if config._log:
                self.log(
                    log_config=self._log_config,
                    prompt=prompt,
                    response=Response(content=content),
                    **log_kwargs,
                )

            return Response(content=content)

    async def achat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[Response, Stream]:
        (
            system_prompt,
            user_prompt,
            all_kwargs,
            log_kwargs,
        ) = self._process_args(prompt, config, **config_kwargs)

        if config.stream:
            output = await self.replicate_client.async_stream(
                config.model,
                input={"prompt": user_prompt, **all_kwargs},
            )

            return Stream(
                iterator=output,
                log_config=self._log_config,
                kwargs=self._get_log_kwargs(prompt, None, **log_kwargs),
                processor=_process_chunk,
            )
        else:
            output = await self.replicate_client.async_run(
                config.model,
                input={"prompt": user_prompt, **all_kwargs},
            )

            content = "".join(item for item in output)

            if config._log:
                self.log(
                    log_config=self._log_config,
                    prompt=prompt,
                    response=Response(content=content),
                    **log_kwargs,
                )

            return Response(content=content)



"""
Name: Fireworks
URL: https://fireworks.ai
Features:
- Chat Completion
- Completion (todo)
"""
from openai import AsyncOpenAI, OpenAI

from .openai import OpenAIConnector


class FireworksConnector(OpenAIConnector):
    def __init__(
        self,
        client: "BaseClient",
        api_key: str,
        log_config: "LogConfig" = None,
        **kwargs
    ):
        super().__init__(client=client, api_key=None, log_config=log_config)
        self.api_key = api_key
        self.client = OpenAI(
            base_url="https://api.fireworks.ai/inference/v1",
            api_key=self.api_key,
            **kwargs
        )
        self.async_client = AsyncOpenAI(
            base_url="https://api.fireworks.ai/inference/v1",
            api_key=self.api_key,
            **kwargs
        )



from typing import Tuple, Union

from .chat.client import IChatClient
from .chat.entities import (
    ChatConfig,
    ChatConfigTypes,
    LogConfig,
    Prompt,
    PromptTypes,
    Response,
    ResponseTypes,
    PromptContext,
)
from .connections._litellm import LiteLLMConnector
from .connections.anthropic import AnthropicConnector
from .connections.fireworks import FireworksConnector
from .connections.openai import OpenAIConnector
from .connections.openai_azure import AzureOpenAIConnector
from .connections.replicate import ReplicateConnector
from .logs.app import app


# Todo: Add BaseClient
# Todo: Add AsyncResource, SyncResource
class BaseClient:
    api_key: Union[str, None]
    api_keys: dict[str, str]
    endpoint: Union[str, None]
    azure_openai_config: dict[str, str]
    debug: bool = False
    ws_endpoint: str

    def __init__(self, debug: bool = False, ws_endpoint: str = None):
        self.debug = debug
        self.ws_endpoint = ws_endpoint or "wss://api.getspeck.ai/debug/ws"
        self.azure_openai_config = {}
        self.prompt = BasePrompt(self)

    def add_api_key(self, provider: str, api_key: str):
        self.api_keys[provider] = api_key

    def add_azure_openai_config(self, azure_endpoint: str, api_version: str):
        self.azure_openai_config = {
            "azure_endpoint": azure_endpoint,
            "api_version": api_version,
        }

    def to_dict(self):
        return {
            "api_key": self.api_key,
            "api_keys": self.api_keys,
            "endpoint": self.endpoint,
            "azure_openai_config": self.azure_openai_config,
            "debug": self.debug,
        }


class Resource:
    pass


class AsyncResource(Resource):
    pass


class SyncResource(Resource):
    pass


class Logger(SyncResource):  # App logger
    def __init__(self, client: BaseClient):
        self.client = client

    def log(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.log(*args, **kwargs)

    def info(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.debug(*args, **kwargs)

    def warning(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.critical(*args, **kwargs)

    def exception(self, *args, **kwargs):
        kwargs["endpoint"] = self.client.endpoint
        app.exception(*args, **kwargs)


def _create_connector(
    client: BaseClient, prompt: PromptTypes, config: ChatConfig = None, **config_kwargs
) -> IChatClient:
    if config is None:
        config = ChatConfig(**config_kwargs)
        # Todo: convert to default config based on class param
    elif len(config_kwargs) > 0:
        # Set config_kwargs as config attributes
        for key, value in config_kwargs.items():
            setattr(config, key, value)

    if config.provider is None:
        # Try to extract provider by getting string before : in model
        if ":" in config.model:
            provider_str, model_str = config.model.split(":", 1)
            config.provider = provider_str
            config.model = model_str
        else:
            raise ValueError("Provider must be specified in config or as a class param")

    provider = config.provider.lower()
    if client.api_keys.get(provider) is None and config.provider != "litellm":
        raise ValueError(f"An API key for {config.provider} is required")

    log_config = LogConfig(
        api_key=config_kwargs.get("_speck_access_token") or client.api_key,
        endpoint=client.endpoint or "https://api.getspeck.ai",
    )
    if "_speck_access_token" in config_kwargs:
        del config_kwargs["_speck_access_token"]

    if provider == "fireworks":
        connector = FireworksConnector(
            client=client,
            api_key=client.api_keys["fireworks"].strip(),
            log_config=log_config,
        )
        return connector
    if provider == "openai":
        connector = OpenAIConnector(
            client=client,
            api_key=client.api_keys["openai"].strip(),
            log_config=log_config,
        )
        return connector
    if provider == "azure-openai":
        connector = AzureOpenAIConnector(
            client=client,
            api_key=client.api_keys["azure-openai"].strip(),
            log_config=log_config,
            **client.azure_openai_config,
        )
        return connector
    if provider == "replicate":
        connector = ReplicateConnector(
            client=client,
            api_key=client.api_keys["replicate"].strip(),
            log_config=log_config,
        )
        return connector
    if provider == "anthropic":
        connector = AnthropicConnector(
            client=client,
            api_key=client.api_keys["anthropic"].strip(),
            log_config=log_config,
        )
        return connector
    if provider == "litellm":
        connector = LiteLLMConnector(
            client=client,
            api_key="",
            log_config=log_config,
        )
        return connector

    # raise ValueError("Provider not found")
    connector = LiteLLMConnector(
        client=client,
        api_key="",
        log_config=log_config,
    )
    return connector


class BasePrompt(SyncResource):
    client: BaseClient

    def __init__(self, client: BaseClient):
        self.client = client

    def _create_context(self, path: str):
        return PromptContext(api_key=self.client.api_key, endpoint=self.client.endpoint, path=path)

    # Deprecated
    # def read_prompt(self, path: str, id: Union[str, None] = None):
    #     with self._create_context() as context:
    #         prompt = Prompt.read(path, id, context=context)
    #     return prompt

    def read_prompts(self, path: str) -> dict[str, Prompt]:
        with self._create_context(path=path) as context:
            prompts = Prompt.read_all(path, context=context)
        return prompts

    def from_string(self, text: str, override: bool = False):
        with self._create_context(path=text.split("\n")[0]) as context:
            prompts = Prompt.reads_all(text, context=context)
        return prompts


class BaseChat(SyncResource):
    client: BaseClient

    def __init__(self, client: BaseClient):
        self.client = client

    def _create_connector(
        self, prompt: PromptTypes, config, **config_kwargs
    ) -> (Prompt, ChatConfig, IChatClient, Union[str, None]):
        _access_token = config_kwargs.get("_speck_access_token")
        if config_kwargs is not None and "_speck_access_token" in config_kwargs:
            del config_kwargs["_speck_access_token"]

        prompt = Prompt.create(prompt)
        config = ChatConfig.create(config, config_kwargs)
        connector = _create_connector(
            self.client, prompt, config, _speck_access_token=_access_token
        )

        if self.client.debug:
            # Create a socket connection to the server
            prompt, config = connector.debug_chat(prompt, config)

        return prompt, config, connector


class Chat(BaseChat):
    def __init__(self, client: BaseClient):
        super().__init__(client)

    def create(
        self, *, prompt: PromptTypes, config: ChatConfig = None, **config_kwargs
    ):
        prompt, config, connector = self._create_connector(
            prompt, config, **config_kwargs
        )
        return connector.chat(prompt=prompt, config=config)

    def log(
        self, messages: PromptTypes, config: ChatConfigTypes, response: ResponseTypes, custom_metadata: dict = None
    ):
        prompt = Prompt.create(messages)
        config = ChatConfig.create(config)
        response = Response.create(response)
        config.log_chat(log_config=LogConfig(api_key=self.client.api_key, endpoint=self.client.endpoint), prompt=prompt, response=response, custom_metadata=custom_metadata)


class AsyncChat(BaseChat):
    def __init__(self, client: BaseClient):
        super().__init__(client)

    async def create(
        self, *, prompt: PromptTypes, config: ChatConfig = None, **config_kwargs
    ):
        prompt, config, connector = self._create_connector(
            prompt, config, **config_kwargs
        )
        return await connector.achat(prompt, config)

    def log(
        self,
        log_config: LogConfig,
        messages: Prompt,
        config: ChatConfig,
        response: Response,
        custom_metadata: dict = None,
    ):
        config.log_chat(log_config=log_config, prompt=messages, response=response, custom_metadata=custom_metadata)


class Speck(BaseClient):
    def __init__(
        self,
        api_key: Union[str, None] = None,
        api_keys: dict[str, str] = {},
        endpoint: str = "https://api.getspeck.ai",
        debug: bool = False,
        **kwargs,
    ):
        super().__init__(debug=debug, **kwargs)
        self.api_key = api_key.strip() if api_key is not None else None
        self.api_keys = api_keys
        self.endpoint = endpoint

        self.chat = Chat(self)
        self.logger = Logger(self)


class AsyncSpeck(BaseClient):
    def __init__(
        self,
        api_key: Union[str, None] = None,
        api_keys: dict[str, str] = {},
        endpoint: Union[str, None] = "https://api.getspeck.ai",
        debug: bool = False,
        **kwargs,
    ):
        super().__init__(debug=debug, **kwargs)
        self.api_key = api_key.strip() if api_key is not None else None
        self.api_keys = api_keys
        self.endpoint = endpoint

        self.chat = AsyncChat(self)
        self.logger = Logger(self)


Client = Speck
AsyncClient = AsyncSpeck



"""
Name: OpenAI Azure
URL:
Features:
"""
from openai import AsyncAzureOpenAI, AzureOpenAI

from .openai import OpenAIConnector


class AzureOpenAIConnector(OpenAIConnector):
    def __init__(
        self,
        client: "BaseClient",
        api_key: str,
        azure_endpoint: str,
        api_version: str,
        log_config: "LogConfig" = None,
        **kwargs
    ):
        super().__init__(client=client, api_key=None, log_config=log_config)
        self.client = AzureOpenAI(
            api_key=api_key,
            azure_endpoint=azure_endpoint,
            api_version=api_version,
            **kwargs
        )
        self.async_client = AsyncAzureOpenAI(
            api_key=api_key,
            azure_endpoint=azure_endpoint,
            api_version=api_version,
            **kwargs
        )
        self.api_key = api_key
        self.azure_endpoint = azure_endpoint
        self.api_version = api_version



from ._openai import OpenAIWrapper as OpenAI
from ._openai import AsyncOpenAIWrapper as AsyncOpenAI
from ._openai import AzureOpenAIWrapper as AzureOpenAI
from ._openai import AsyncAzureOpenAIWrapper as AsyncAzureOpenAI



from setuptools import find_packages, setup


def readme():
    with open("README.md") as f:
        README = f.read()
    return README


def parse_requirements(filename):
    with open(filename, "r") as f:
        requirements = f.read().splitlines()
    return requirements


install_requires = [
    "openai",
    "replicate",
    "requests",
    "pydantic",
    "httpx",
    "websockets",
    "litellm",
]


setup(
    name="speck",
    packages=find_packages(exclude=["tests", "tests.*"]),
    version="0.1.4",
    description="Speck - Development and observability toolkit for LLM apps.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords=[
        "speck",
        "openai",
        "llm",
        "ai",
        "chat",
        "bot",
        "gpt",
        "gpt-3",
        "gpt-4",
        "anthropic",
        "replicate",
        "litellm",
        "observability",
    ],
    url="https://github.com/speckai/speck",
    download_url="https://github.com/speckai/speck/archive/refs/tags/v0.1.4.tar.gz",
    homepage="https://getspeck.ai",
    author="",
    author_email="Lucas Jaggernauth <luke@getspeck.ai>, Raghav Pillai <raghav@getspeck.ai>",
    classifiers=[
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    install_requires=install_requires,
)



from typing import Union

from litellm import acompletion, completion

from ..chat.entities import NOT_GIVEN, ChatConfig, Prompt, Stream
from ..util._filter import no_filter_kwargs

"""
Name: OpenAI Azure
URL:
Features:
"""
from openai import AsyncAzureOpenAI, AzureOpenAI

from .openai import OpenAIConnector, OpenAIResponse, _process_chunk


class LiteLLMClient:
    def __init__(self, method):
        self.method = method
        self.chat = self.Chat(self)

    class Chat:
        def __init__(self, client):
            self.client = client
            self.completions = self.Completions(self.client)

        class Completions:
            def __init__(self, client):
                self.client = client

            def create(self, *args, **kwargs):
                return self.client.method(*args, **kwargs)


class LiteLLMConnector(OpenAIConnector):
    def __init__(
        self,
        client: "BaseClient",
        api_key: str,
        log_config: "LogConfig" = None,
        **kwargs
    ):
        super().__init__(client=client, api_key=None, log_config=log_config)
        self.client = LiteLLMClient(method=completion, **kwargs)
        self.async_client = LiteLLMClient(method=acompletion, **kwargs)
        # self.api_key = api_key

    def chat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[OpenAIResponse, Stream]:
        input, all_kwargs = self._process_kwargs(prompt, config, **config_kwargs)

        if config.stream:
            output_stream = self.client.chat.completions.create(
                messages=input,
                **no_filter_kwargs(completion, all_kwargs, self.EXCLUDED_PARAMS),
            )

            return Stream(
                client=self._client,
                iterator=output_stream,
                log_config=self._log_config,
                kwargs=self._get_log_kwargs(prompt, None, **all_kwargs),
                processor=_process_chunk,
            )
        else:
            output = self.client.chat.completions.create(
                messages=input,
                **no_filter_kwargs(completion, all_kwargs, self.EXCLUDED_PARAMS),
            )

            if config._log:
                self.log(
                    log_config=self._log_config,
                    prompt=prompt,
                    response=OpenAIResponse(output),
                    **all_kwargs,
                )
                # Todo: set config= as param

        return OpenAIResponse(output)

    async def achat(
        self, prompt: Prompt, config: ChatConfig = NOT_GIVEN, **config_kwargs
    ) -> Union[OpenAIResponse, Stream]:
        input, all_kwargs = self._process_kwargs(prompt, config, **config_kwargs)

        if config.stream:
            output_stream = await self.async_client.chat.completions.create(
                messages=input,
                **no_filter_kwargs(acompletion, all_kwargs, self.EXCLUDED_PARAMS),
            )

            # Todo: async iterator support
            return Stream(
                client=self._client,
                iterator=output_stream,
                log_config=self._log_config,
                kwargs=self._get_log_kwargs(prompt, None, **all_kwargs),
                processor=_process_chunk,
            )
        else:
            output = await self.async_client.chat.completions.create(
                messages=input,
                **no_filter_kwargs(acompletion, all_kwargs, self.EXCLUDED_PARAMS),
            )

            if config._log:
                self.log(
                    log_config=self._log_config,
                    prompt=prompt,
                    response=OpenAIResponse(output),
                    **all_kwargs,
                )
                # Todo: set config= as param

        return OpenAIResponse(output)



from ._client import AsyncClient, AsyncSpeck, Client, Speck
from .chat import *
from .connections import *
from .logs import logger
from .wrapper import OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI



from __future__ import annotations

import logging
import os
from typing import Any, Callable, Iterator

from openai import OpenAI as _OpenAI
from openai import AsyncOpenAI as _AsyncOpenAI
from openai import AzureOpenAI as _AzureOpenAI
from openai import AsyncAzureOpenAI as _AsyncAzureOpenAI
from openai import Stream as _Stream
from litellm import completion

from .. import ChatLogger, OpenAIResponse, ProvidersList, Speck
from .._client import BasePrompt
from ..chat.entities import (
    ChatConfig,
    LogConfig,
    OpenAIChatConfig,
    Prompt,
    Response,
    Stream,
)
from ..util._wrapper import wrap_method

logger = logging.getLogger(__name__)


def wrapper(original_method, *args, **kwargs):
    """
    Example of a wrapper function that can be used with add_method_wrapper.

    Args:
        original_method: The original method to be wrapped.
        *args, **kwargs: Arguments and keyword arguments for the original method.

    Returns:
        The result of the original method call.
    """
    logger.info(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"After calling {original_method.__name__}")
    return result


class OpenAIStream:
    # processor that has lambda which returns MessageDelta
    def __init__(
        self,
        config: OpenAIChatConfig,
        prompt: Prompt,
        iterator: Iterator[Any],
        kwargs: dict,
        processor: Callable[[Any], Any],
        speck_api_key: str,
        provider: str = "openai",
        prompt_ref: dict = None,
        custom_metadata: dict = None,
        _speck_client: Speck = None,
    ):
        self.config: OpenAIChatConfig = config
        self.prompt = prompt
        self.message: str = ""
        self._iterator = iterator
        self._kwargs = kwargs
        self._processor = processor
        self._has_logged = False
        self._closed = False
        self._last_item = None
        self.speck_api_key = speck_api_key
        self.provider = provider
        self.prompt_ref = prompt_ref
        self._speck_client = _speck_client
        self.custom_metadata = custom_metadata

    def _log(self):
        if not self._has_logged:
            self._has_logged = True

            kwargs = self._kwargs
            kwargs["prompt"] = self._kwargs.get("prompt", [])
            kwargs["model"] = self._kwargs.get("model", "N/A")
            # kwargs["response"] = Response(content=self.message, raw={}, closed=True)
            # ChatLogger.log(**kwargs)

            self.config.log_chat(
                log_config=LogConfig(api_key=self.speck_api_key, endpoint=self._speck_client.endpoint),
                prompt=self.prompt,
                response=Response(content=self.message, raw={}, closed=True),
                provider=self.provider,
                prompt_ref=self.prompt_ref,
                custom_metadata=self.custom_metadata,
            )

    def _process(self, item) -> Any:
        return self._processor(item)

    def __next__(self) -> Any:
        try:
            if self._closed:
                raise StopIteration

            item: Any = self._process(next(self._iterator))
            self._last_item = item
            if item.choices[0].delta.content:
                self.message += item.choices[0].delta.content
            return item
        except StopIteration:
            self._log()
            raise

    def __iter__(self) -> Iterator[Any]:
        return self

    def close(self):
        try:
            self._closed = True
            # todo: make this work for packages other than openai
            self._iterator.response.close()
        except AttributeError:
            pass

    async def __anext__(self) -> Any:
        try:
            if self._closed:
                raise StopAsyncIteration

            item: Any = self._process(await self._iterator.__anext__())
            self._last_item = item
            if item.choices[0].delta.content:
                self.message += item.choices[0].delta.content
            return item
        except StopAsyncIteration:
            self._log()
            raise

    def __aiter__(self) -> Iterator[Any]:
        return self


def preprocess_chat(**kwargs):
    model: str = kwargs.get("model", None)
    stream = kwargs.get("stream", False)
    messages = kwargs.get("messages", None)
    if isinstance(messages, Prompt):
        prompt = messages
        kwargs["messages"] = prompt.compose()
    else:
        prompt = Prompt.from_openai(messages)

    custom_metadata = kwargs.get("metadata", None)
    if "metadata" in kwargs:
        del kwargs["metadata"]

    config = OpenAIChatConfig(**kwargs)
    config = config.convert()

    return prompt, config, stream, model, custom_metadata, kwargs

def chat_wrapper(self: OpenAIWrapper, original_method, *args, **kwargs):
    prompt, config, stream, model, custom_metadata, kwargs = preprocess_chat(**kwargs)
    # if model is not None and ":" in model:
    #     provider, model = model.split(":", 1)
    #     if provider in ProvidersList:
    #         config.provider = provider
    #         config.model = model
    #         # Todo: return in OpenAI format
    #         return self._speck_client.chat.create(prompt=prompt, config=config)
    logger.info(f"Call {original_method.__name__} with model {model}")
    provider = "fireworks" if self.base_url is not None and "fireworks" in str(self.base_url.raw.raw_host) else "openai"
    config.provider = provider

    if stream:
        stream = original_method(*args, **kwargs)
        # Todo: wrap the Stream class
        # Best current way to do this is to convert our Stream class to an OpenAI Stream class
        return OpenAIStream(
            config=config,
            prompt=prompt,
            iterator=stream,
            kwargs={
                "provider": provider,
            },
            processor=lambda a: a,
            speck_api_key=self._speck_api_key,
            provider=provider,
            prompt_ref=prompt.get_ref(),
            custom_metadata=custom_metadata,
            _speck_client=self._speck_client,
        )
    else:
        result = original_method(*args, **kwargs)
        config.log_chat(
            log_config=LogConfig(api_key=self._speck_api_key),
            prompt=prompt,
            response=OpenAIResponse(result),
            provider=provider,
            prompt_ref=prompt.get_ref(),
            custom_metadata=custom_metadata,
        )

        return result


async def async_chat_wrapper(self: AsyncOpenAIWrapper, original_method, *args, **kwargs):
    prompt, config, stream, model, custom_metadata, kwargs = preprocess_chat(**kwargs)
    # if model is not None and ":" in model:
    #     provider, model = model.split(":", 1)
    #     if provider in ProvidersList:
    #         config.provider = provider
    #         config.model = model
    #         # Todo: return in OpenAI format
    #         return self._speck_client.chat.create(prompt=prompt, config=config)
    logger.info(f"Call {original_method.__name__} with model {model}")
    provider = "fireworks" if self.base_url is not None and "fireworks" in str(self.base_url.raw.raw_host) else "openai"
    config.provider = provider

    if stream:
        stream = await original_method(*args, **kwargs)
        # Todo: wrap the Stream class
        # Best current way to do this is to convert our Stream class to an OpenAI Stream class
        return OpenAIStream(
            config=config,
            prompt=prompt,
            iterator=stream,
            kwargs={
                "provider": provider,
            },
            processor=lambda a: a,
            speck_api_key=self._speck_api_key,
            provider=provider,
            prompt_ref=prompt.get_ref(),
            custom_metadata=custom_metadata,
            _speck_client=self._speck_client,
        )
    else:
        result = await original_method(*args, **kwargs)
        config.log_chat(
            log_config=LogConfig(api_key=self._speck_api_key),
            prompt=prompt,
            response=OpenAIResponse(result),
            provider=provider,
            prompt_ref=prompt.get_ref(),
            custom_metadata=custom_metadata,
        )

        return result


def stream_next_wrapper(original_method, *args, **kwargs):
    logger.info(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"After calling {original_method.__name__}")
    return result


def stream_iter_wrapper(original_method, *args, **kwargs):
    logger.info(f"Before calling {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"After calling {original_method.__name__}")
    return result


class OpenAIWrapper(_OpenAI):
    _speck_api_key: str = None
    _speck_client: Speck = None
    prompt: BasePrompt

    def __init__(self, *args, **kwargs):
        if "speck_api_key" in kwargs:
            del kwargs["speck_api_key"]
        super().__init__(*args, **kwargs)

    def initialize_speck(
        self,
        api_key: str = None,
        speck_api_key: str = None,
        api_keys: dict[str, str] = {},
        debug: bool = False,
    ):
        """
        Deprecated. Replaced with:
        openai = OpenAI(..., speck_api_key=...)
        """
        self._speck_api_key = api_key or speck_api_key
        if "openai" not in api_keys:
            api_keys["openai"] = self.api_key
        if api_keys["openai"] is None:
            api_keys["openai"] = os.environ.get("OPENAI_API_KEY")

        self._speck_client = Speck(
            api_key=self._speck_api_key, api_keys=api_keys, debug=debug
        )

    def add_api_key(self, provider: str, api_key: str):
        self._speck_client.add_api_key(provider, api_key)

    def add_api_keys(self, api_keys: dict[str, str]):
        for provider, api_key in api_keys.items():
            self.add_api_key(provider, api_key)


class AzureOpenAIWrapper(_AzureOpenAI):
    _speck_api_key: str = None
    _speck_client: Speck = None
    prompt: BasePrompt

    def __init__(self, *args, **kwargs):
        if "speck_api_key" in kwargs:
            del kwargs["speck_api_key"]
        super().__init__(*args, **kwargs)

    def add_api_key(self, provider: str, api_key: str):
        self._speck_client.add_api_key(provider, api_key)

    def add_api_keys(self, api_keys: dict[str, str]):
        for provider, api_key in api_keys.items():
            self.add_api_key(provider, api_key)


class AsyncOpenAIWrapper(_AsyncOpenAI):
    _speck_api_key: str = None
    _speck_client: Speck = None
    prompt: BasePrompt

    def __init__(self, *args, **kwargs):
        if "speck_api_key" in kwargs:
            del kwargs["speck_api_key"]
        super().__init__(*args, **kwargs)

    def add_api_key(self, provider: str, api_key: str):
        self._speck_client.add_api_key(provider, api_key)

    def add_api_keys(self, api_keys: dict[str, str]):
        for provider, api_key in api_keys.items():
            self.add_api_key(provider, api_key)


class AsyncAzureOpenAIWrapper(_AsyncAzureOpenAI):
    _speck_api_key: str = None
    _speck_client: Speck = None
    prompt: BasePrompt

    def __init__(self, *args, **kwargs):
        if "speck_api_key" in kwargs:
            del kwargs["speck_api_key"]
        super().__init__(*args, **kwargs)

    def add_api_key(self, provider: str, api_key: str):
        self._speck_client.add_api_key(provider, api_key)

    def add_api_keys(self, api_keys: dict[str, str]):
        for provider, api_key in api_keys.items():
            self.add_api_key(provider, api_key)


def _wrapper_init(original_method, *args, **kwargs):
    logger.info(f"Initializing {original_method.__name__}")
    result = original_method(*args, **kwargs)
    logger.info(f"Adding method wrappers {original_method.__name__}")
    self = args[0]
    self._speck_api_key = kwargs.get("speck_api_key", None) or os.environ.get("SPECK_API_KEY", None)
    if "speck_api_key" in kwargs:
        del kwargs["speck_api_key"]

    self._speck_client = Speck(api_key=self._speck_api_key)
    self.prompt = BasePrompt(self._speck_client)

    def chat_wrapper_caller(*args, **kwargs):
        return chat_wrapper(self,*args, **kwargs)

    async def async_chat_wrapper_caller(*args, **kwargs):
        return await async_chat_wrapper(self, *args, **kwargs)

    is_sync = isinstance(self, OpenAIWrapper) or isinstance(self, AzureOpenAIWrapper)
    wrap_method(
        self.chat.completions,
        "create",
        chat_wrapper_caller if is_sync else async_chat_wrapper_caller,
    )
    logger.info(f"After calling {original_method.__name__}")
    return result


wrap_method(OpenAIWrapper, "__init__", _wrapper_init)
wrap_method(AsyncOpenAIWrapper, "__init__", _wrapper_init)
wrap_method(AzureOpenAIWrapper, "__init__", _wrapper_init)
wrap_method(AsyncAzureOpenAIWrapper, "__init__", _wrapper_init)

# add_method_kwarg(OpenAIWrapper, "__init__", "speck_log", 69)
